@@ -9,6 +9,7 @@
 """Helpers for managing NSS databases in Directory Server
 """
 
+import copy
 import os
 import re
 import socket
@@ -18,12 +19,20 @@ import logging
 # from nss import nss
 import subprocess
 from datetime import datetime, timedelta
-from subprocess import check_output
+from subprocess import check_output, run, PIPE
 from lib389.passwd import password_generate
-
+from lib389._mapped_object_lint import DSLint
+from lib389.lint import DSCERTLE0001, DSCERTLE0002
 from lib389.utils import ensure_str, format_cmd_list
 import uuid
 
+# Setuptools ships with 'packaging' module, let's use it from there
+try:
+    from pkg_resources.extern.packaging.version import LegacyVersion
+# Fallback to a normal 'packaging' module in case 'setuptools' is stripped
+except:
+    from packaging.version import LegacyVersion
+
 KEYBITS = 4096
 CA_NAME = 'Self-Signed-CA'
 CERT_NAME = 'Server-Cert'
@@ -41,7 +50,7 @@ VALID_MIN = 61  # Days
 log = logging.getLogger(__name__)
 
 
-class NssSsl(object):
+class NssSsl(DSLint):
     def __init__(self, dirsrv=None, dbpassword=None, dbpath=None):
         self.dirsrv = dirsrv
         self._certdb = dbpath
@@ -55,9 +64,37 @@ class NssSsl(object):
         else:
             self.dbpassword = dbpassword
 
-        self.db_files = {"dbm_backend": ["%s/%s" % (self._certdb, f) for f in ("key3.db", "cert8.db", "secmod.db")],
-                         "sql_backend": ["%s/%s" % (self._certdb, f) for f in ("key4.db", "cert9.db", "pkcs11.txt")],
-                         "support": ["%s/%s" % (self._certdb, f) for f in ("noise.txt", PIN_TXT, PWD_TXT)]}
+        self.db_files = {group: [f"{self._certdb}/{f}" for f in files]
+                         for group, files in {"dbm_backend": ("key3.db", "cert8.db", "secmod.db"),
+                                              "sql_backend": ("key4.db", "cert9.db", "pkcs11.txt"),
+                                              "support": ("noise.txt", PIN_TXT, PWD_TXT)}.items()}
+
+    @classmethod
+    def lint_uid(cls):
+        return 'ssl'
+
+    def _lint_certificate_expiration(self):
+        """Check all the certificates in the db if they will expire within 30 days
+        or have already expired.
+        """
+        cert_list = []
+        all_certs = self._rsa_cert_list()
+        for cert in all_certs:
+            cert_list.append(self.get_cert_details(cert[0]))
+
+        for cert in cert_list:
+            cert_date = cert[3].split()[0]
+            diff_date = datetime.strptime(cert_date, '%Y-%m-%d').date() - datetime.today().date()
+            if diff_date < timedelta(days=0):
+                # Expired
+                report = copy.deepcopy(DSCERTLE0002)
+                report['detail'] = report['detail'].replace('CERT', cert[0])
+                yield report
+            elif diff_date < timedelta(days=30):
+                # Expiring within 30 days
+                report = copy.deepcopy(DSCERTLE0001)
+                report['detail'] = report['detail'].replace('CERT', cert[0])
+                yield report
 
     def detect_alt_names(self, alt_names=[]):
         """Attempt to determine appropriate subject alternate names for a host.
@@ -120,18 +157,19 @@ class NssSsl(object):
         except FileExistsError:
             pass
 
-        # Write a README to let people know what this is
-        readme_file = '%s/%s' % (self._certdb, 'README.txt')
-        if not os.path.exists(readme_file):
-            with open(readme_file, 'w') as f:
-                f.write("""
+        if self.dirsrv is None:
+            # Write a README to let people know what this is
+            readme_file = '%s/%s' % (self._certdb, 'README.txt')
+            if not os.path.exists(readme_file):
+                with open(readme_file, 'w') as f:
+                    f.write("""
 SSCA - Simple Self-Signed Certificate Authority
 
 This is part of the 389 Directory Server project's lib389 toolkit. It
 creates a simple, standalone certificate authority for testing and
 development purposes. It's suitable for evaluation and testing purposes
 only.
-                """)
+                    """)
 
         # In the future we may add the needed option to avoid writing the pin
         # files.
@@ -184,6 +222,24 @@ only.
         assert not self._db_exists()
         return True
 
+    def openssl_rehash(self, certdir):
+        """
+        Compatibly run c_rehash (on old openssl versions) or openssl rehash (on
+        new ones). Prefers openssl rehash, because openssl on versions where
+        the rehash command doesn't exist, also doesn't correctly set the return
+        code. Instead, we parse the output of `openssl version` and try to
+        figure out if we have a new enough version to unconditionally run rehash.
+        """
+        openssl_version = check_output(['/usr/bin/openssl', 'version']).decode('utf-8').strip()
+        rehash_available = LegacyVersion(openssl_version.split(' ')[1]) >= LegacyVersion('1.1.0')
+
+        if rehash_available:
+            cmd = ['/usr/bin/openssl', 'rehash', certdir]
+        else:
+            cmd = ['/usr/bin/c_rehash', certdir]
+        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
+        check_output(cmd, stderr=subprocess.STDOUT)
+
     def create_rsa_ca(self, months=VALID):
         """
         Create a self signed CA.
@@ -208,6 +264,7 @@ only.
             'CT,,',
             '-v',
             '%s' % months,
+            '-2',
             '--keyUsage',
             'certSigning',
             '-d',
@@ -217,8 +274,9 @@ only.
             '-f',
             '%s/%s' % (self._certdb, PWD_TXT),
         ]
+        cmd_input = b'y\n\n'  # responses to certutil questions
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        result = ensure_str(run(cmd, check=True, stderr=PIPE, stdout=PIPE, input=cmd_input).stdout)
         self.log.debug("nss output: %s", result)
         # Now extract the CAcert to a well know place.
         # This allows us to point the cacert dir here and it "just works"
@@ -235,9 +293,7 @@ only.
         certdetails = check_output(cmd, stderr=subprocess.STDOUT)
         with open('%s/ca.crt' % self._certdb, 'w') as f:
             f.write(ensure_str(certdetails))
-        cmd = ['/usr/bin/c_rehash', self._certdb]
-        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        self.openssl_rehash(self._certdb)
         return True
 
     def rsa_ca_needs_renew(self):
@@ -317,9 +373,7 @@ only.
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
         check_output(cmd, stderr=subprocess.STDOUT)
 
-        cmd = ['/usr/bin/c_rehash', self._certdb]
-        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        self.openssl_rehash(self._certdb)
 
         # Import the new CA to our DB instead of the old CA
         cmd = [
@@ -361,6 +415,9 @@ only.
         for line in lines:
             if line == '':
                 continue
+            if line == 'Database needs user init':
+                # There are no certs, abort...
+                return []
             cert_values.append(re.match(r'^(.+[^\s])[\s]+([^\s]+)$', line.rstrip()).groups())
         return cert_values
 
@@ -572,9 +629,7 @@ only.
 
         if ca is not None:
             shutil.copyfile(ca, '%s/ca.crt' % self._certdb)
-            cmd = ['/usr/bin/c_rehash', self._certdb]
-            self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-            check_output(cmd, stderr=subprocess.STDOUT)
+            self.openssl_rehash(self._certdb)
             cmd = [
                 '/usr/bin/certutil',
                 '-A',
@@ -929,5 +984,3 @@ only.
             # Remove the p12
             if os.path.exists(p12_bundle):
                 os.remove(p12_bundle)
-
-
